@@ -5,7 +5,7 @@ use std::{
 };
 
 use exocore_chain::{block::BlockOffset, operation::OperationId};
-use exocore_core::time::Clock;
+use exocore_core::time::{Clock, ConsistentTimestamp};
 use exocore_protos::{core::EntityGarbageCollectorConfig, store::EntityMutation};
 
 use super::{sort_mutations_commit_time, EntityAggregator};
@@ -52,6 +52,7 @@ pub struct GarbageCollector {
     config: GarbageCollectorConfig,
     max_entity_deleted_duration: chrono::Duration,
     max_trait_deleted_duration: chrono::Duration,
+    trait_versions_min_age: chrono::Duration,
     clock: Clock,
     inner: Arc<RwLock<Inner>>,
 }
@@ -66,6 +67,9 @@ impl GarbageCollector {
             chrono::Duration::from_std(config.deleted_trait_collection)
                 .expect("Couldn't convert `deleted_trait_collection` to chrono::Duration");
 
+        let trait_versions_min_age = chrono::Duration::from_std(config.min_operation_age)
+            .expect("Couldn't convert `trait_versions_min_age` to chrono::Duration");
+
         let inner = Arc::new(RwLock::new(Inner {
             config,
             operations: Vec::with_capacity(config.queue_size),
@@ -76,6 +80,7 @@ impl GarbageCollector {
             config,
             max_entity_deleted_duration,
             max_trait_deleted_duration,
+            trait_versions_min_age,
             clock,
             inner,
         }
@@ -83,24 +88,30 @@ impl GarbageCollector {
 
     /// Checks if an entity for which we collected its mutation metadata from
     /// the index should be added to the garbage collection queue.
-    pub fn maybe_flag_for_collection(&self, entity_id: &str, aggregator: &EntityAggregator) {
-        let last_block_offset = if let Some(offset) = aggregator.last_block_offset {
-            offset
-        } else {
-            // we don't collect if operations are only in pending
+    pub fn maybe_flag_for_collection(&self, aggregator: &EntityAggregator) {
+        if aggregator.in_pending || aggregator.pending_deletion {
+            // we don't collect if any of the operations are still in pending store or if
+            // a deletion is pending for the entity
             return;
-        };
+        }
+
+        let last_block_offset = aggregator
+            .last_block_offset
+            .expect("No last_block_offset, but no operations in pending");
 
         let now = self.clock.now_chrono();
 
         if let Some(deletion_date) = &aggregator.deletion_date {
             let elapsed = now.signed_duration_since(*deletion_date);
             if elapsed >= self.max_entity_deleted_duration {
-                trace!("Collecting entity {} since got deleted since", entity_id);
+                trace!(
+                    "Collecting entity {} since got deleted since",
+                    aggregator.entity_id
+                );
                 let mut inner = self.inner.write().expect("Fail to acquire inner lock");
                 inner.maybe_enqueue(Operation::DeleteEntity(
                     last_block_offset,
-                    entity_id.to_string(),
+                    aggregator.entity_id.to_string(),
                 ));
                 return;
             }
@@ -112,13 +123,13 @@ impl GarbageCollector {
                 if elapsed >= self.max_trait_deleted_duration {
                     trace!(
                         "Collecting entity {} trait {} since it got deleted",
-                        entity_id,
+                        aggregator.entity_id,
                         trait_id
                     );
                     let mut inner = self.inner.write().expect("Fail to acquire inner lock");
                     inner.maybe_enqueue(Operation::DeleteTrait(
                         last_block_offset,
-                        entity_id.to_string(),
+                        aggregator.entity_id.to_string(),
                         trait_id.clone(),
                     ));
                     continue;
@@ -129,13 +140,13 @@ impl GarbageCollector {
                 let mut inner = self.inner.write().expect("Fail to acquire inner lock");
                 trace!(
                     "Collecting entity {} trait {} since it has too many versions ({})",
-                    entity_id,
+                    aggregator.entity_id,
                     trait_id,
                     trait_aggr.mutation_count,
                 );
                 inner.maybe_enqueue(Operation::CompactTrait(
                     last_block_offset,
-                    entity_id.to_string(),
+                    aggregator.entity_id.to_string(),
                     trait_id.clone(),
                 ));
             }
@@ -162,6 +173,8 @@ impl GarbageCollector {
             return Vec::new();
         }
 
+        let min_op_time = self.clock.now_chrono() - self.trait_versions_min_age;
+
         debug!(
             "Starting a garbage collection pass with {} operations for {} entities...",
             operations.len(),
@@ -173,13 +186,16 @@ impl GarbageCollector {
             let entity_id = op.entity_id();
 
             // get all mutations for entity until block offset where we deemed the entity to
-            // be collectable
+            // be collectable and until minimum operation age.
             let mutations = if let Ok(mut_res) = entity_fetcher(entity_id) {
-                let mutations =
-                    mut_res
-                        .mutations
-                        .into_iter()
-                        .filter(|mutation| matches!(mutation.block_offset, Some(offset) if offset <= until_block_offset));
+                let mutations = mut_res.mutations.into_iter().take_while(|mutation| {
+                    let op_time = ConsistentTimestamp::from(mutation.operation_id).to_datetime();
+                    if op_time > min_op_time {
+                        return false;
+                    }
+
+                    matches!(mutation.block_offset, Some(offset) if offset <= until_block_offset)
+                });
 
                 sort_mutations_commit_time(mutations)
             } else {
@@ -206,11 +222,14 @@ impl GarbageCollector {
                 deletions.push(deletion);
             }
         }
-        info!(
-            "Garbage collection generated {} deletion operations for {} entities",
-            deletions.len(),
-            entity_count,
-        );
+
+        if !deletions.is_empty() {
+            info!(
+                "Garbage collection generated {} deletion operations for {} entities",
+                deletions.len(),
+                entity_count,
+            );
+        }
 
         deletions
     }
@@ -231,6 +250,9 @@ pub struct GarbageCollectorConfig {
     /// After how many versions for a trait a compaction is triggered.
     pub trait_versions_leeway: usize,
 
+    /// Minimum age an operation needs to have in order to be collected.
+    pub min_operation_age: Duration,
+
     /// Size of the queue of entities to be collected.
     pub queue_size: usize,
 }
@@ -242,7 +264,8 @@ impl Default for GarbageCollectorConfig {
             deleted_trait_collection: Duration::from_secs(14 * DAY_SECS),
             trait_versions_max: 5,
             trait_versions_leeway: 7,
-            queue_size: 100,
+            min_operation_age: Duration::from_secs(7 * DAY_SECS),
+            queue_size: 300,
         }
     }
 }
@@ -348,17 +371,10 @@ fn filter_trait_mutations<'i, I>(
 where
     I: Iterator<Item = MutationMetadata> + 'i,
 {
+    use crate::local::mutation_index::MutationType;
     mutations.filter(move |mutation| match &mutation.mutation_type {
-        crate::local::mutation_index::MutationType::TraitPut(put_mut)
-            if put_mut.trait_id == trait_id =>
-        {
-            true
-        }
-        crate::local::mutation_index::MutationType::TraitTombstone(tomb_trait_id)
-            if tomb_trait_id == trait_id =>
-        {
-            true
-        }
+        MutationType::TraitPut(put_mut) if put_mut.trait_id == trait_id => true,
+        MutationType::TraitTombstone(tomb_trait_id) if tomb_trait_id == trait_id => true,
         _ => false,
     })
 }
@@ -426,7 +442,10 @@ mod tests {
         *,
     };
     use crate::{
-        local::mutation_index::{MutationType, PutTraitMetadata},
+        local::{
+            entity_index::aggregator::tests::mock_pending_delete,
+            mutation_index::{MutationType, PutTraitMetadata},
+        },
         ordering::OrderingValueWrapper,
     };
 
@@ -438,6 +457,7 @@ mod tests {
 
         let config = GarbageCollectorConfig {
             deleted_entity_collection: Duration::from_secs(1),
+            min_operation_age: Duration::from_secs(1),
             ..Default::default()
         };
         let gc = GarbageCollector::new(config, clock.clone());
@@ -452,7 +472,7 @@ mod tests {
 
         {
             // entity cannot be collected since deletion date is not 1 sec later
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 0);
         }
 
@@ -460,7 +480,7 @@ mod tests {
         clock.add_fixed_instant_duration(Duration::from_secs(2));
 
         {
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 1);
 
             // operation considered not in chain, shouldn't get deleted
@@ -476,7 +496,7 @@ mod tests {
         }
 
         {
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 1);
 
             // operation considered now in chain, should get deleted
@@ -501,6 +521,7 @@ mod tests {
 
         let config = GarbageCollectorConfig {
             deleted_trait_collection: Duration::from_secs(1),
+            min_operation_age: Duration::from_secs(1),
             ..Default::default()
         };
         let gc = GarbageCollector::new(config, clock.clone());
@@ -517,7 +538,7 @@ mod tests {
 
         {
             // entity cannot be collected since deletion date is not 1 sec later
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 0);
         }
 
@@ -525,7 +546,7 @@ mod tests {
         clock.add_fixed_instant_duration(Duration::from_secs(2));
 
         {
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 1);
 
             // operation considered not in chain, shouldn't get deleted
@@ -541,7 +562,7 @@ mod tests {
         }
 
         {
-            gc.maybe_flag_for_collection("et1", &aggregator);
+            gc.maybe_flag_for_collection(&aggregator);
             assert_queue_len(&gc, 1);
 
             // operation considered now in chain, should get deleted
@@ -578,16 +599,20 @@ mod tests {
         ];
         let aggregator = EntityAggregator::new(mutations.into_iter()).unwrap();
 
+        // move time a second later
+        clock.add_fixed_instant_duration(Duration::from_secs(2));
+
         // not enough versions to be collected
         let gc = GarbageCollector::new(
             GarbageCollectorConfig {
                 trait_versions_max: 3,
                 trait_versions_leeway: 6,
+                min_operation_age: Duration::from_secs(1),
                 ..Default::default()
             },
             clock.clone(),
         );
-        gc.maybe_flag_for_collection("et1", &aggregator);
+        gc.maybe_flag_for_collection(&aggregator);
         assert_queue_len(&gc, 0);
 
         // should now have enough versions to be collected
@@ -595,11 +620,12 @@ mod tests {
             GarbageCollectorConfig {
                 trait_versions_max: 3,
                 trait_versions_leeway: 4,
+                min_operation_age: Duration::from_secs(1),
                 ..Default::default()
             },
             clock,
         );
-        gc.maybe_flag_for_collection("et1", &aggregator);
+        gc.maybe_flag_for_collection(&aggregator);
         assert_queue_len(&gc, 1);
 
         // two first trait should get deleted
@@ -618,6 +644,77 @@ mod tests {
         assert_eq!(extract_ops(deletions), vec![put1_op, put2_op]);
     }
 
+    #[test]
+    fn cannot_collect_if_pending_deletion() {
+        let node = LocalNode::generate();
+        let now = Instant::now() - Duration::from_secs(60);
+        let clock = Clock::new_fixed_mocked(now);
+
+        let put1_op: u64 = clock.consistent_time(node.node()).into();
+        let put2_op: u64 = put1_op + 1;
+        let put3_op: u64 = put2_op + 1;
+        let put4_op: u64 = put3_op + 1;
+        let put5_op: u64 = put4_op + 1;
+        let mutations = vec![
+            mock_put_trait("trt1", "typ", Some(2), put1_op, None, None),
+            mock_put_trait("trt1", "typ", Some(3), put2_op, None, None),
+            mock_put_trait("trt1", "typ", Some(4), put3_op, None, None),
+            mock_put_trait("trt1", "typ", Some(5), put4_op, None, None),
+            mock_pending_delete(Some(6), put5_op),
+        ];
+        let aggregator = EntityAggregator::new(mutations.into_iter()).unwrap();
+
+        // should be collecting, but one pending deletion, so won't collect
+        let gc = GarbageCollector::new(
+            GarbageCollectorConfig {
+                trait_versions_max: 2,
+                trait_versions_leeway: 3,
+                min_operation_age: Duration::from_secs(1),
+                ..Default::default()
+            },
+            clock,
+        );
+        gc.maybe_flag_for_collection(&aggregator);
+        assert_queue_len(&gc, 0);
+    }
+
+    #[test]
+    fn cannot_collect_if_any_operations_in_pending() {
+        let node = LocalNode::generate();
+        let now = Instant::now() - Duration::from_secs(60);
+        let clock = Clock::new_fixed_mocked(now);
+
+        let put1_op: u64 = clock.consistent_time(node.node()).into();
+        let put2_op: u64 = put1_op + 1;
+        let put3_op: u64 = put2_op + 1;
+        let put4_op: u64 = put3_op + 1;
+        let put5_op: u64 = put4_op + 1;
+        let mutations = vec![
+            mock_put_trait("trt1", "typ", Some(2), put1_op, None, None),
+            mock_put_trait("trt1", "typ", Some(3), put2_op, None, None),
+            mock_put_trait("trt1", "typ", Some(4), put3_op, None, None),
+            mock_put_trait("trt1", "typ", Some(5), put4_op, None, None),
+            mock_put_trait(
+                "trt1", "typ", None, /* in pending */
+                put5_op, None, None,
+            ),
+        ];
+        let aggregator = EntityAggregator::new(mutations.into_iter()).unwrap();
+
+        // should be collecting, but one operation in pending, so won't collect
+        let gc = GarbageCollector::new(
+            GarbageCollectorConfig {
+                trait_versions_max: 2,
+                trait_versions_leeway: 3,
+                min_operation_age: Duration::from_secs(1),
+                ..Default::default()
+            },
+            clock,
+        );
+        gc.maybe_flag_for_collection(&aggregator);
+        assert_queue_len(&gc, 0);
+    }
+
     fn assert_queue_len(gc: &GarbageCollector, len: usize) {
         let inner = gc.inner.read().unwrap();
         assert_eq!(inner.operations.len(), len);
