@@ -3,10 +3,7 @@ use std::{
     collections::{HashMap, HashSet},
     path::{Path, PathBuf},
     rc::Rc,
-    sync::{
-        atomic::{AtomicU64, Ordering},
-        Arc,
-    },
+    sync::Arc,
     time::Instant,
 };
 
@@ -75,7 +72,6 @@ where
     full_cell: FullCell,
     chain_handle: EngineHandle<CS, PS>,
     gc: GarbageCollector,
-    gc_last_run_block_height: AtomicU64,
 }
 
 impl<CS, PS> EntityIndex<CS, PS>
@@ -117,7 +113,6 @@ where
             full_cell: cell,
             chain_handle,
             gc: GarbageCollector::new(config.garbage_collector, clock),
-            gc_last_run_block_height: AtomicU64::new(0),
         };
 
         let chain_last_block = index.chain_handle.get_chain_last_block_info()?;
@@ -167,7 +162,7 @@ where
                 affected_operations.push(op_id);
                 continue;
             } else if !batched_operations.is_empty() {
-                let current_operations = std::mem::replace(&mut batched_operations, Vec::new());
+                let current_operations = std::mem::take(&mut batched_operations);
                 index_operations_count +=
                     self.handle_chain_pending_operations(current_operations.into_iter())?;
             }
@@ -331,13 +326,12 @@ where
                     // 
                     // We let the garbage collector know that the entity may need to be garbage collected. The
                     // actual garbage collection will be done asynchronously at later time.
-                    self.gc.maybe_flag_for_collection(&entity_id, &entity_mutations);
+                    self.gc.maybe_flag_for_collection(&entity_mutations);
 
                     if !query_include_deleted {
                         return None;
                     }
                 }
-
                 matched_entities.insert(matched_mutation.entity_id.clone());
 
                 // TODO: Support for negative re-scoring https://github.com/appaquet/exocore/issues/143
@@ -456,48 +450,21 @@ where
     }
 
     /// Calls the garbage collector to run a pass on entities that got flagged
-    /// to be collector.
+    /// and generates deletion mutations.
     pub fn run_garbage_collector(&self) -> Result<Vec<EntityMutation>, Error> {
         let last_chain_indexed_block = self
             .last_chain_indexed_block()
             .map_err(|err| anyhow!("Couldn't get last chain indexed block: {}", err))?;
-        let (_, last_chain_index_height) = if let Some(info) = last_chain_indexed_block {
-            info
-        } else {
+        if last_chain_indexed_block.is_none() {
             // we can only run GC from chain index
             return Ok(Vec::new());
         };
 
-        let last_run_block_height = self.gc_last_run_block_height.load(Ordering::Acquire);
-        if last_run_block_height == 0 {
-            // GC didn't run since start of process. We wait for next chain re-indexation to
-            // make sure they aren't deletion mutations that haven't been
-            // applied to chain index yet because of min depth constraint.
-            self.gc_last_run_block_height
-                .store(last_chain_index_height, Ordering::Release);
-            return Ok(Vec::new());
-        }
-
-        let min_depth = self
-            .config
-            .chain_index_min_depth
-            .max(self.config.chain_index_depth_leeway);
-        if last_chain_index_height - last_run_block_height < min_depth {
-            // deletion mutations that may have been generated by last GC may not have been
-            // applied to the chain index yet. we need to run for them to be applied before
-            // running again so that we don't re-collect them again.
-            return Ok(Vec::new());
-        }
-
-        let mutations = self
+        let deletions = self
             .gc
             .run(|entity_id| self.chain_index.fetch_entity_mutations(entity_id));
-        if !mutations.is_empty() {
-            self.gc_last_run_block_height
-                .store(last_chain_index_height, Ordering::Release);
-        }
 
-        Ok(mutations)
+        Ok(deletions)
     }
 
     /// Creates the chain index based on configuration.
@@ -667,10 +634,15 @@ where
                     && last_chain_block_height.saturating_sub(*height) >= chain_index_min_depth
             })
             .flat_map(|(offset, _height, engine_operation)| {
+                let operation_id = engine_operation.operation_id;
+                let (index_ops, entity_id) =
+                    IndexOperation::from_chain_engine_operation(engine_operation, offset);
+
                 if !pending_index_empty {
                     // delete from pending index if it's not already empty
-                    pending_index_mutations.push(IndexOperation::DeleteOperation(
-                        engine_operation.operation_id,
+                    pending_index_mutations.push(IndexOperation::DeleteEntityOperation(
+                        entity_id,
+                        operation_id,
                     ));
                 }
 
@@ -679,7 +651,7 @@ where
                     new_highest_block_offset = Some(offset);
                 }
 
-                IndexOperation::from_chain_engine_operation(engine_operation, offset)
+                index_ops
             });
 
         self.chain_index.apply_operations(chain_index_mutations)?;
@@ -793,6 +765,10 @@ where
         include_deleted: bool,
     ) {
         for entity_result in entity_results {
+            if entity_result.mutations.should_collect() {
+                self.gc.maybe_flag_for_collection(&entity_result.mutations);
+            }
+
             let traits = self.fetch_entity_traits(&entity_result.mutations, include_deleted);
             if let Some(entity) = entity_result.proto.entity.as_mut() {
                 entity.traits = traits;
@@ -809,7 +785,7 @@ where
         entity_mutations
             .traits
             .iter()
-            .flat_map(|(trait_id, agg)| {
+            .filter_map(|(trait_id, agg)| {
                 if let Some(projection) = &agg.projection {
                     if projection.skip {
                         return None;
@@ -837,6 +813,14 @@ where
                     }
                 };
 
+                if mutation.entity_id != entity_mutations.entity_id {
+                    error!(
+                        "Fetched from chain operation {} that didn't belong to entity {}, but entity {}",
+                        mut_metadata.operation_id, entity_mutations.entity_id, mutation.entity_id
+                    );
+                    return None;
+                }
+
                 let mut trt = match mutation.mutation? {
                     Mutation::PutTrait(put_mut) => put_mut.r#trait,
                     Mutation::DeleteTrait(_)
